import os
import json

from tqdm.auto import tqdm
import numpy as np

import torch
import torch.nn as nn



def sinusoidal_embedding(n, d):
    '''Create sinusoidal embeddings.'''
    return torch.tensor([[np.sin(i / 10000 ** (2 * j / d)) 
                          if j % 2 == 0 else np.cos(i / 10000 ** (2 * j / d))
                          for j in range(d)] for i in range(n)],
                          dtype=torch.float32)

class ConditionalDiffusionModel(nn.Module):
    '''A denoise diffusion model.'''
    def __init__(self, n_feature, 
                 n_time_feature, 
                 condition_classes,
                 n_condition_feature,
                 n_steps=1000, 
                 min_beta=1e-4, 
                 max_beta=0.02,
                 n_network_blocks=3,
                 ):
        super(ConditionalDiffusionModel, self).__init__()
        self.n_feature = n_feature
        self.n_steps = n_steps
        self.betas =  nn.Parameter(torch.linspace(min_beta, max_beta, n_steps),
                                   requires_grad=False)
        self.alphas = nn.Parameter(1 - self.betas, requires_grad=False)
        self.alpha_bars = nn.Parameter(
                            torch.tensor([torch.prod(self.alphas[:i + 1])
                            for i in range(len(self.alphas))]),
                            requires_grad=False)
        
        # Define the temporal embeddings
        self.time_embeddings = nn.Parameter(
            sinusoidal_embedding(n_steps, n_time_feature), requires_grad=False)
        
        # Define the condition embeddings
        self.condition_classes = list(np.unique(condition_classes))
        self.condition_to_idx = {c: i for i, c in 
                                 enumerate(self.condition_classes)}
        self.condition_embeddings = nn.Parameter(1e-3 * 
            torch.randn(len(self.condition_classes), n_condition_feature),
            requires_grad=True)

        # Define the denoiser
        self.network = MLPDenoiser(n_feature, 
                                   n_time_feature,
                                   n_condition_feature,
                                   n_blocks=n_network_blocks,
                                   )
        
    def forward(self, x0, t, eta=None):
        '''Add noise to the data.
        x0 has shape (n_batch, n_feature) and t has shape (n_batch,).
        '''
        device = next(self.parameters()).device

        n_batch, n_feature = x0.shape
        x0 = x0.to(device)
        a_bar = self.alpha_bars[t]

        if eta is None:
            eta = torch.randn(n_batch, n_feature).to(device)

        noisy = (a_bar.sqrt().reshape(n_batch, 1) * x0 
                 + (1 - a_bar).sqrt().reshape(n_batch, 1) * eta)

        return noisy
    
    def backward(self, x, t, x_cond):
        '''Predict the mean of distribution at t-1 given the data at t.'''
        t_emb = self.time_embeddings[t.reshape(-1)]
        return self.network(x, t_emb, x_cond)
    
    def generate(self, x_cond):
        '''Generate samples from the model.'''
        device = next(self.parameters()).device

        n_feature = self.n_feature
        n_sample = len(x_cond)

        # Start from random noise
        x = torch.randn(n_sample, n_feature).to(device)

        with torch.no_grad():
            for idx in range(self.n_steps):
                t = self.n_steps - idx - 1
                # Estimate teh noise to remove
                time_tensor = (torch.ones(n_sample,) * t).long().to(device)
                eta_theta = self.backward(x, time_tensor, x_cond)

                alpha_t = self.alphas[t]
                alpha_t_bar = self.alpha_bars[t]

                # Partially denoise the data
                x = (1 / alpha_t.sqrt()) * (x 
                        - (1 - alpha_t) / (1 - alpha_t_bar).sqrt() * eta_theta)

                # Add noise except for the last step
                if t > 0:
                    z = torch.randn(n_sample, n_feature).to(device)

                    # Define the noise level
                    beta_t = self.betas[t]
                    sigma_t = beta_t.sqrt()

                    # Add noise
                    x = x + sigma_t * z

        return x

class RecurrentBlock(nn.Module):
    '''A MLP style recurrent block.'''
    def __init__(self, n_input, n_output, n_hidden):
        super(RecurrentBlock, self).__init__()
        self.fc1 = nn.Linear(n_input, n_hidden)
        self.act = nn.LeakyReLU()
        self.fc2 = nn.Linear(n_hidden, n_output)
        self.norm = nn.LayerNorm(n_hidden)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.fc2(x)
        x = self.norm(x)
        return x
    
class MLPDenoiser(nn.Module):
    '''A simple MLP denoiser.'''
    def __init__(self, n_feature, n_time_feature, n_condition_feature,
                 n_hidden=2048, n_blocks=3):
        super(MLPDenoiser, self).__init__()

        self.fc1 = nn.Linear(n_feature + n_time_feature + n_condition_feature, 
                             n_hidden)

        self.blocks = nn.ModuleList()
        for _ in range(n_blocks):
            self.blocks.append(RecurrentBlock(
                n_feature + n_time_feature + n_condition_feature + n_hidden,
                n_hidden,
                n_hidden))

        self.fc2 = nn.Linear(
            n_feature + n_time_feature + n_condition_feature + n_hidden, 
            n_feature)

    def forward(self, x, t_emb, x_cond):
        x_hidden = self.fc1(torch.cat([x, t_emb, x_cond], dim=1))

        for block in self.blocks:
            x_hidden = block(torch.cat([x, t_emb, x_cond, x_hidden], dim=1))

        x_out = self.fc2(torch.cat([x, t_emb, x_cond, x_hidden], dim=1))
        return x_out
    
def train_diffusion_model(
        model, 
        loader, 
        num_epochs, 
        output_path,
        lr=1e-4,
        loss_history=None,
    ):
    '''Train the diffusion model.'''
    os.makedirs(output_path, exist_ok=True)
    device = next(model.parameters()).device

    if loss_history is None:
        loss_history = {'loss': []}

    optimizer = torch.optim.AdamW(model.parameters(), lr=lr)
    mse = nn.MSELoss()
    best_loss = float('inf')

    pbar = tqdm(range(num_epochs))
    for epoch in pbar:
        epoch_loss = 0.0

        for batch in loader:
            # Load data
            cond_classes = batch['cell_type']
            x0 = batch['X_ce_latent'].to(device)
            n = len(x0)

            # Pick a noise and a time step for each data point
            eta = torch.randn_like(x0).to(device)
            t = torch.randint(0, model.n_steps, (n,)).to(device)

            # Add noise
            noisy_x = model(x0, t, eta)

            # Predict the noise to remove
            eta_theta = model.backward(noisy_x, t, cond_classes)

            # Compute the loss
            loss = mse(eta_theta, eta)
            optimizer.zero_grad()
            loss.backward()
            nn.utils.clip_grad_norm_(model.parameters(), 2)
            optimizer.step()

            epoch_loss += loss.item() * len(x0) / len(loader.dataset)

        loss_history['loss'].append(epoch_loss)
        pbar.set_description(f"Loss: {epoch_loss}")

        with open(f'{output_path}/loss_history.json', 'w') as f:
            json.dump(loss_history, f)

        # Store the model if it is the best
        if best_loss > epoch_loss:
            best_loss = epoch_loss
            torch.save(model, f'{output_path}/model.pt')
            torch.save(model.state_dict(), f'{output_path}/best_state_dict.pth')


def generate_cells(model, cond_classes, batch_size=512):
    '''Generate cells from the model.'''
    n_sample = len(cond_classes)
    n_batches = n_sample // batch_size + 1

    samples = []
    with torch.no_grad():
        for idx in tqdm(range(n_batches)):
            start_idx = idx * batch_size
            end_idx = min(n_sample, (idx + 1) * batch_size)

            cond_indices = [model.condition_to_idx[c] for c in cond_classes[start_idx:end_idx]]
            x_cond = model.condition_embeddings[cond_indices]
            batch_cells = model.generate(x_cond)
            samples.append(batch_cells.detach().cpu().numpy())

    return np.concatenate(samples, axis=0)

def generate_transition_cells(model, start_cell_type, end_cell_type, n_cells,
                              batch_size=512):
    '''Generate cells between two cell types.'''
    start_ct_emb = model.condition_embeddings[model.condition_to_idx[start_cell_type]].reshape(1, -1)
    end_ct_emb = model.condition_embeddings[model.condition_to_idx[end_cell_type]].reshape(1, -1)

    emb_weights = torch.linspace(0, 1, n_cells).reshape(-1, 1).to(start_ct_emb.device)
    n_batches = n_cells // batch_size + 1

    samples = []
    with torch.no_grad():
        for idx in tqdm(range(n_batches)):
            start_idx = idx * batch_size
            end_idx = min(n_cells, (idx + 1) * batch_size)

            batch_weights = emb_weights[start_idx:end_idx]
            batch_cells = model.generate(
                batch_weights * end_ct_emb + (1 - batch_weights) * start_ct_emb)
            samples.append(batch_cells.detach().cpu().numpy())

    cond_classes = [start_cell_type] * (n_cells // 2) + [end_cell_type] * (n_cells - (n_cells // 2))

    return np.concatenate(samples, axis=0), cond_classes




import torch
import torch.nn as nn
import torch.nn.functional as F


class MLP(nn.Module):
    def __init__(self,
                 input_dim,
                 output_dim,
                 hidden_dims=(256,),
                 dropout_prob=0.05,
                ):
        super().__init__()
        
        self.layers = nn.ModuleList()
        prev_dim = input_dim
        
        for hidden_dim in hidden_dims:
            self.layers.append(nn.Linear(prev_dim, hidden_dim))
            self.layers.append(nn.LayerNorm(hidden_dim))
            self.layers.append(nn.LeakyReLU())
            self.layers.append(nn.Dropout(dropout_prob))
            prev_dim = hidden_dim
            
        self.layers.append(nn.Linear(prev_dim, output_dim))
                
    def forward(self, X):        
        for layer in self.layers:
            X = layer(X)
            
        return X


import os
import json

from tqdm.auto import tqdm
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

from scmg.model.basic import MLP

torch.set_float32_matmul_precision('high')



def edge_batch_to_Xs(batch, exp_data, dataset_to_cell_idx_map,
                      max_dropout_rate=0.3):

    cell_ref_idx = batch['cell_ref_idx']
    cell_query_idx = batch['cell_query_idx']
    cell_query_datasets = batch['dataset_query']
    
    # Sample the negative cells
    neg_cell_idx = []
    for ref_cell_dataset in batch['dataset_ref']:
        rand_idx = np.random.randint(low=0,
                    high=len(dataset_to_cell_idx_map[ref_cell_dataset]))
        neg_cell_idx.append(dataset_to_cell_idx_map[ref_cell_dataset][rand_idx])

    X_anchor = exp_data[cell_ref_idx]['X_exp']
    X_positive = exp_data[cell_query_idx]['X_exp']
    X_negative = exp_data[neg_cell_idx]['X_exp']

    # Randomly drop out some of the input data
    dropout_rate = np.random.uniform(low=0, high=1)
    X_anchor_with_drop = X_anchor
    X_positive_with_drop = X_positive
    X_negative_with_drop = X_negative
    
    if dropout_rate < max_dropout_rate:
        X_anchor_with_drop = X_anchor * (torch.rand(*X_anchor.shape) > dropout_rate)
        X_positive_with_drop = X_positive * (torch.rand(*X_anchor.shape) > dropout_rate)
        X_negative_with_drop = X_negative * (torch.rand(*X_anchor.shape) > dropout_rate)

    return X_anchor_with_drop, X_positive_with_drop, X_negative_with_drop, X_positive, cell_query_datasets

class CellEmbedder(nn.Module):
    def __init__(self,
                 n_genes,
                 dataset_id_map,
                 ):
        super(CellEmbedder, self).__init__()
        self.dataset_id_map = dataset_id_map
        self.dataset_emb = nn.parameter.Parameter(
            data=torch.randn(len(dataset_id_map), 64), requires_grad=True)

        self.encoder = MLP(n_genes, 512, hidden_dims=(2048, 2048,),
                           dropout_prob=0.0)
        self.decoder = MLP(512 + 64, n_genes, hidden_dims=(1024, 2048,),
                           dropout_prob=0)
    
    def forward(self, X_exp):
        return self.encoder(X_exp)
    
    def decode(self, X_emb, dataset_names):
        dataset_ids = [self.dataset_id_map[dn] for dn in dataset_names]
        dataset_emb = self.dataset_emb[dataset_ids]
        X_emb_cat = torch.cat([X_emb, dataset_emb], dim=1)

        return F.softplus(self.decoder(X_emb_cat))
        
def l2_loss(X_emb):
    return torch.mean(torch.sum(X_emb.pow(2), dim=1))

def contrastive_loss(X_emb_anchor, X_emb_pos, X_emb_neg):
    # Calculate the similarity between the anchor and the neighbors
    dists_pos = torch.sqrt(torch.sum((X_emb_anchor - X_emb_pos).pow(2), 
                                     dim=1) + 1e-8)
    dists_neg = torch.sqrt(torch.sum((X_emb_anchor - X_emb_neg).pow(2), 
                                     dim=1) + 1e-8)
    dist_diff = dists_pos - dists_neg

    # Calculate the contrastive loss
    loss = F.softplus(dist_diff) 
    return torch.mean(loss)

def recon_loss(X_exp, X_pred):
    return torch.sum(torch.square(X_pred - X_exp)) / X_exp.shape[0]

def calc_loss(model, batch, exp_data, dataset_to_cell_idx_map, device,
            contrastive_weight=1000, l2_weight=1e-1, recon_weight=1e-3):
    (X_anchor_with_drop, X_positive_with_drop, X_negative_with_drop, 
     X_positive, cell_query_datasets) = edge_batch_to_Xs(
                                batch, exp_data, dataset_to_cell_idx_map)
    
    X_anchor_with_drop = X_anchor_with_drop.to(device)
    X_positive_with_drop = X_positive_with_drop.to(device)
    X_negative_with_drop = X_negative_with_drop.to(device)
    X_positive = X_positive.to(device)

    X_emb_anchor = model(X_anchor_with_drop)
    X_emb_pos = model(X_positive_with_drop)
    X_emb_neg = model(X_negative_with_drop)

    l2 = l2_loss(X_emb_anchor) + l2_loss(X_emb_pos) + l2_loss(X_emb_neg)
    contrastive1 = contrastive_loss(X_emb_anchor, X_emb_pos, X_emb_neg)
    contrastive2 = contrastive_loss(X_emb_pos, X_emb_anchor, X_emb_neg)
    contrastive = contrastive1 + contrastive2

    X_pos_pred = model.decode(X_emb_pos, cell_query_datasets)
    recon = recon_loss(X_positive, X_pos_pred)

    return (contrastive_weight * contrastive,
            l2_weight * l2, recon_weight * recon)

def train_contrastive_embedder(
        model,
        edge_loader,
        exp_data,
        dataset_to_cell_idx_map,
        num_epochs,
        output_path,
        lr=1e-3,
        loss_history=None,
    ):
    os.makedirs(output_path, exist_ok=True)
    device = next(model.parameters()).device

    if loss_history is None:
        loss_history = {'train_contrastive': [], 'train_l2': [],
                        'train_total': [], 'train_recon': []}

    optimizer = torch.optim.AdamW(model.parameters(), lr=lr)

    pbar = tqdm(range(num_epochs))

    for epoch in pbar:

        for k in loss_history:
            loss_history[k].append([])

        model.train()

        for batch in edge_loader:
            model.zero_grad()
            contrastive, l2, recon = calc_loss(model, batch, exp_data, 
                                   dataset_to_cell_idx_map, device)
            loss = contrastive + l2 + recon

            loss.backward()
            nn.utils.clip_grad_norm_(model.parameters(), 2)

            optimizer.step()
            loss_history['train_contrastive'][-1].append(contrastive.item())
            loss_history['train_l2'][-1].append(l2.item())
            loss_history['train_recon'][-1].append(recon.item())
            loss_history['train_total'][-1].append(loss.item())


        mean_loss = np.mean(loss_history['train_total'][-1])
        pbar.set_description(f"Train loss: {mean_loss}")

        with open(f'{output_path}/loss_history.json', 'w') as f:
            json.dump(loss_history, f)
            torch.save(model, f'{output_path}/model.pt')
            torch.save(model.state_dict(), f'{output_path}/best_state_dict.pth')


import scipy.sparse
import pandas as pd
import pkg_resources
import anndata
import sklearn.neighbors


def get_Xs_from_anndata(adata, standard_genes):

    # Define the relationships between indices
    standard_genes = list(standard_genes)
    adata_genes = list(adata.var.index)
    common_ids = np.intersect1d(standard_genes, adata_genes)
    common_in_standard_indices = [standard_genes.index(g) for g in common_ids]
    common_in_adata_indices = [adata_genes.index(g) for g in common_ids]

    # Convert the expression data to numpy array
    if scipy.sparse.issparse(adata.X):
        adata_X = adata.X.toarray()
    else:
        adata_X = adata.X
    
    # Extract the expressions
    X = np.zeros((adata.shape[0], len(standard_genes)), dtype=np.float32)
    X[:, common_in_standard_indices] = adata_X[:, common_in_adata_indices]

    # Normalize the expressions
    X = X / (X.sum(axis=1, keepdims=True) + 1e-6) * 1e4
    X = np.log1p(X)

    # Extract the measurement mask
    X_mask = np.zeros((adata.shape[0], len(standard_genes)), dtype=bool)
    X_mask[:, common_in_standard_indices] = True

    return X, X_mask

def embed_adata(model, 
                 adata, 
                 standard_gene_csv=None,
                 batch_size=512,
                 inplace=True
                 ):
    '''Embed the data in an AnnData object using the given model.
    The adata object should have the following attributes:
    - adata.X should be raw counts.
    - adata.var.index should be human gene Ensembl IDs.
    '''
    device = next(model.parameters()).device

    # Load the standard genes
    if standard_gene_csv is None:
        standard_gene_csv = pkg_resources.resource_filename('scmg',
                            'data/standard_genes.csv')
        
    standard_genes = pd.read_csv(standard_gene_csv)['human_id'].values
    adata_cg = adata[:, adata.var.index.isin(standard_genes)].copy()

    N_cells = adata.shape[0]
    N_batches = int(np.ceil(N_cells / batch_size))
    Z = []

    for i in tqdm(range(N_batches)):
        start = i * batch_size
        stop = min((i + 1) * batch_size, N_cells)
        adata_batch = adata_cg[start:stop].copy()

        X, _ = get_Xs_from_anndata(adata_batch, standard_genes)
        X = torch.tensor(X, dtype=torch.float32).to(device)

        with torch.no_grad():
            z = model(X)
            z = z.detach().cpu().numpy()
            for i in range(z.shape[0]):
                Z.append(z[i])

    if inplace:
        adata.obsm['X_ce_latent'] = np.array(Z)
    else:
        return np.array(Z)

def embed_standardized_adata(model,
                              adata,
                              batch_size=512,
                              inplace=True,
                              verbose_level=1
                              ):
    ''' Embed the data in an AnnData object using the given model. The adata
    should be already standardized. The adata object should have the following
    attributes:
    - adata.X should be raw counts.
    - adata.var.index should be the standarded gene set with the same order.
    - adata.layers['measure_mask'] should be the measurement mask.
    '''
    device = next(model.parameters()).device

    N_cells = adata.shape[0]
    N_batches = int(np.ceil(N_cells / batch_size))
    Z = []

    # Decide what to print based on the verbose level
    if verbose_level == 0:
        iterator = range(N_batches)
    else:
        iterator = tqdm(range(N_batches))

    for i in iterator:
        start = i * batch_size
        stop = min((i + 1) * batch_size, N_cells)
        adata_batch = adata[start:stop].copy()

        X = adata_batch.X
        if scipy.sparse.issparse(X):
            X = X.toarray()

        X = X / (X.sum(axis=1, keepdims=True) + 1e-6) * 1e4
        X = np.log1p(X)

        X = torch.tensor(X, dtype=torch.float32).to(device)
        
        with torch.no_grad():
            z = model(X)
            z = z.detach().cpu().numpy()
            for i in range(z.shape[0]):
                Z.append(z[i])

    if inplace:
        adata.obsm['X_ce_latent'] = np.array(Z)
    else:
        return np.array(Z)
    
def decode_adata(model,
                 adata,
                 dataset_names,
                 standard_gene_csv=None,
                 batch_size=512,):
    '''Decode the data in an AnnData object using the given model.
    The input adata object should have adata.obsm['X_ce_latent'].
    Returns a new AnnData object with the decoded gene expression.
    '''
    device = next(model.parameters()).device
    
    # Load the standard genes
    if standard_gene_csv is None:
        standard_gene_csv = pkg_resources.resource_filename('scmg',
                            'data/standard_genes.csv')
    
    standard_genes = pd.read_csv(standard_gene_csv)['human_id'].values

    N_cells = adata.shape[0]
    N_batches = int(np.ceil(N_cells / batch_size))
    X_pred = np.zeros((N_cells, len(standard_genes)), dtype=np.float32)

    for i in tqdm(range(N_batches)):
        start = i * batch_size
        stop = min((i + 1) * batch_size, N_cells)
        Z_batch = adata.obsm['X_ce_latent'][start:stop].copy()
        dataset_names_batch = dataset_names[start:stop]

        X_pred_batch = model.decode(
            torch.tensor(Z_batch, dtype=torch.float32).to(device),
            dataset_names_batch
            ).detach().cpu().numpy()
        X_pred[start:stop] = X_pred_batch

    adata_pred = anndata.AnnData(
        X = X_pred,
        obs = adata.obs.copy(),
        var = pd.DataFrame(index=standard_genes),
        uns = adata.uns.copy(),
        obsm = adata.obsm.copy(),
        obsp = adata.obsp.copy(),
        )
    
    return adata_pred

def decode_cell_state_embedding(
        model,
        X_ce_latent,
        dataset_names,
        standard_gene_csv=None,
        batch_size=512,
):
    '''Decode the given cell state embeddings using the given model.
    Returns an AnnData object with the decoded gene expression.
    '''
    device = next(model.parameters()).device

    # Load the standard genes
    if standard_gene_csv is None:
        standard_gene_csv = pkg_resources.resource_filename('scmg',
                            'data/standard_genes.csv')
    
    standard_genes = pd.read_csv(standard_gene_csv)['human_id'].values

    N_cells = X_ce_latent.shape[0]
    N_batches = int(np.ceil(N_cells / batch_size))
    X_pred = np.zeros((N_cells, len(standard_genes)), dtype=np.float32)

    for i in tqdm(range(N_batches)):
        start = i * batch_size
        stop = min((i + 1) * batch_size, N_cells)
        Z_batch = X_ce_latent[start:stop].copy()
        dataset_names_batch = dataset_names[start:stop]

        X_pred_batch = model.decode(
            torch.tensor(Z_batch, dtype=torch.float32).to(device),
            dataset_names_batch
            ).detach().cpu().numpy()
        X_pred[start:stop] = X_pred_batch

    adata_pred = anndata.AnnData(
        X = X_pred,
        obs = pd.DataFrame(index=np.arange(N_cells).astype(str)),
        var = pd.DataFrame(index=standard_genes),
        obsm = {'X_ce_latent': X_ce_latent},
        )
    
    return adata_pred

def score_marker_genes(model_ce, adata_std_query, target_emb):
    '''Calculate the marker genes that are most important for determining
    the similarity between the query and target datasets.
    Args:
        model_ce: torch.nn.Module
            The trained contrastive embedding model.
        adata_std_query: anndata.AnnData
            The query dataset with the standard gene order and
            raw expression values.
        target_emb: np.ndarray of shape (n_cells, n_features)
            The embedding of the target dataset.
    '''
    neigh = sklearn.neighbors.NearestNeighbors(n_neighbors=1)
    neigh.fit(target_emb)

    embed_standardized_adata(model_ce, adata_std_query)
    dist0 = np.mean(neigh.kneighbors(adata_std_query.obsm['X_ce_latent'], 
                                     return_distance=True)[0])   
    
    dist_dict = {
        'gene': [],
        'dist_shift': []
    }

    # Calcualte the embedding distance shift after
    # masking out each gene
    for i in tqdm(range(adata_std_query.shape[1])):
        gene = adata_std_query.var.index[i]

        adata_query_tmp = adata_std_query.copy()

        if adata_query_tmp.X[:, i].max() == 0:
            dist = dist0
        else:
            adata_query_tmp.X[:, i] = 0
            embed_standardized_adata(model_ce, adata_query_tmp, verbose_level=0)

            dist = np.mean(neigh.kneighbors(adata_query_tmp.obsm['X_ce_latent'], 
                                 return_distance=True)[0])

        dist_dict['gene'].append(gene)
        dist_dict['dist_shift'].append(dist - dist0)

    dist_df = pd.DataFrame(dist_dict).set_index('gene')
    dist_df['dist_shift_z'] = (dist_df['dist_shift'] - dist_df['dist_shift'].mean()
                               ) / dist_df['dist_shift'].std()
    return dist_df


import warnings

from tqdm import tqdm
import numpy as np
import pandas as pd
import scipy.stats
import scipy.spatial

from sklearn.gaussian_process import GaussianProcessRegressor
import sklearn.gaussian_process.kernels as kernels



class CausalGenePredictor:
    def __init__(self, adata_pert, gene_stds):
        '''
        Args:
            adata_pert (AnnData): Annotated data object with the perturbation
                shifts.
        '''
        self.standard_genes = adata_pert.var.index.values
        self.gene_stds = np.maximum(gene_stds, 0.1)

        self.adata_pert = adata_pert[
            adata_pert.obs['perturbed_gene'].isin(self.standard_genes)].copy()
        
    def calc_causal_scores(self, mean_exp_shifts):
        '''
        Args:
            exp_shifts (np.ndarray): The mean expression shifts for the cell
                state transition.

        Returns:
            pd.DataFrame: A DataFrame with the causal scores for each perturbation.
        '''
        mean_shift_z = mean_exp_shifts / self.gene_stds

        # Calculate the match score for each perturbation
        pert_match_df = self.adata_pert.obs.copy()
        gene_z_shifts = []
        pert_sims = []

        for i in tqdm(range(self.adata_pert.X.shape[0])):
            pert_vec = self.adata_pert.X[i]
            pert_vec_z = pert_vec / self.gene_stds

            if (np.std(pert_vec) == 0) or (np.std(mean_shift_z) == 0):
                pms = 0
            else:
                pms = 1 - scipy.spatial.distance.cosine(
                    mean_shift_z, pert_vec_z)

            pert_gene = self.adata_pert.obs['perturbed_gene'].iloc[i]
            pert_gene_i = self.adata_pert.var.index.get_loc(pert_gene)

            gene_z_shifts.append(mean_shift_z[pert_gene_i])
            pert_sims.append(pms)

        pert_match_df['gene_shift_z'] = np.clip(gene_z_shifts, -5, 5)
        pert_match_df['pert_sim'] = pert_sims
        pert_match_df['pert_match_score'] = pert_sims * pert_match_df['perturbation_sign']
        pert_match_df['causal_score'] = pert_match_df['pert_match_score'] * pert_match_df['gene_shift_z']

        return pert_match_df
        


import json
import warnings
import pkg_resources

import numpy as np
import pandas as pd
import scipy.sparse

import anndata
import scanpy as sc


class GeneNameMapper:
    '''A class to map gene names and IDs for mouse and human.'''
    def __init__(self, 
                 human_gene_csv=None, 
                 mouse_gene_csv=None, 
                 human_mouse_orthologues_csv=None, 
                 mouse_human_orthologues_csv=None,
                 ):
        
        # Define the input files
        if human_gene_csv is None:
            human_gene_csv = pkg_resources.resource_filename('scmg', 
                            'data/gene_names/human_genes.csv')
        if mouse_gene_csv is None:
            mouse_gene_csv = pkg_resources.resource_filename('scmg', 
                            'data/gene_names/mouse_genes.csv')
        if human_mouse_orthologues_csv is None:
            human_mouse_orthologues_csv = pkg_resources.resource_filename('scmg', 
                            'data/gene_names/orthologue_map_human2mouse_best.csv')
        if mouse_human_orthologues_csv is None:
            mouse_human_orthologues_csv = pkg_resources.resource_filename('scmg',
                            'data/gene_names/orthologue_map_mouse2human_best.csv')
    
        # Load the data frames
        human_gene_df = pd.read_csv(human_gene_csv, index_col=0)
        human_gene_df = human_gene_df[~human_gene_df['gene_name'].isna()]
        
        mouse_gene_df = pd.read_csv(mouse_gene_csv, index_col=0)
        mouse_gene_df = mouse_gene_df[~mouse_gene_df['gene_name'].isna()]
        
        human_mouse_df = pd.read_csv(human_mouse_orthologues_csv, index_col=0)
        mouse_human_df = pd.read_csv(mouse_human_orthologues_csv, index_col=0)
        
        # Create dictionaries for mapping
        self.d_human_id_to_name = {human_gene_df.index[i] : human_gene_df['gene_name'].iloc[i]
                                  for i in range(human_gene_df.shape[0])}
        self.d_human_name_to_id = {human_gene_df['gene_name'].iloc[i] : human_gene_df.index[i]
                                  for i in range(human_gene_df.shape[0])}
        
        self.d_mouse_id_to_name = {mouse_gene_df.index[i] : mouse_gene_df['gene_name'].iloc[i]
                                  for i in range(mouse_gene_df.shape[0])}
        self.d_mouse_name_to_id = {mouse_gene_df['gene_name'].iloc[i] : mouse_gene_df.index[i]
                                  for i in range(mouse_gene_df.shape[0])}
        
        self.d_orth_human_to_mouse = {human_mouse_df.index[i] : human_mouse_df['mus_musculus'].iloc[i]
                                     for i in range(human_mouse_df.shape[0])}
        self.d_orth_mouse_to_human = {mouse_human_df.index[i] : mouse_human_df['homo_sapiens'].iloc[i]
                                     for i in range(mouse_human_df.shape[0])}
        
    def map_gene_names(self, genes, source_organism, target_organism, source_type, target_type):
        '''Map gene names or IDs.
        Args:
            genes: The list of genes to be mapped.
            source_organism: The organism of the source genes. Choose from ("human", "mouse").
            traget_organism: The organism of the target genes. Choose from ("human", "mouse").
            source_type: The type of source genes. Choose from ("id", "name").
            target_type: The type of target genes. Choose from ("id", "name").
        '''
        if (source_organism == target_organism) and (source_type == target_type):
            return genes
        
        elif ((source_organism == 'human') and (target_organism == 'human') 
           and (source_type == 'id') and (target_type == 'name')):
            return [self.d_human_id_to_name.get(g, 'na') for g in genes]
        
        elif ((source_organism == 'human') and (target_organism == 'human') 
           and (source_type == 'name') and (target_type == 'id')):
            return [self.d_human_name_to_id.get(g, 'na') for g in genes]
        
        elif ((source_organism == 'mouse') and (target_organism == 'mouse') 
           and (source_type == 'id') and (target_type == 'name')):
            return [self.d_mouse_id_to_name.get(g, 'na') for g in genes]
        
        elif ((source_organism == 'mouse') and (target_organism == 'mouse') 
           and (source_type == 'name') and (target_type == 'id')):
            return [self.d_mouse_name_to_id.get(g, 'na') for g in genes]
        
        elif ((source_organism == 'mouse') and (target_organism == 'human') 
           and (source_type == 'id') and (target_type == 'id')):
            return [self.d_orth_mouse_to_human.get(g, 'na') for g in genes]
        
        elif ((source_organism == 'mouse') and (target_organism == 'human') 
           and (source_type == 'name') and (target_type == 'id')):
            genes = [self.d_mouse_name_to_id.get(g, 'na') for g in genes]
            return [self.d_orth_mouse_to_human.get(g, 'na') for g in genes]
        
        elif ((source_organism == 'mouse') and (target_organism == 'human') 
           and (source_type == 'id') and (target_type == 'name')):
            genes = [self.d_orth_mouse_to_human.get(g, 'na') for g in genes]
            return [self.d_human_id_to_name.get(g, 'na') for g in genes]
        
        elif ((source_organism == 'mouse') and (target_organism == 'human') 
           and (source_type == 'name') and (target_type == 'name')):
            genes = [self.d_mouse_name_to_id.get(g, 'na') for g in genes]
            genes = [self.d_orth_mouse_to_human.get(g, 'na') for g in genes]
            return [self.d_human_id_to_name.get(g, 'na') for g in genes]
        
        elif ((source_organism == 'human') and (target_organism == 'mouse') 
           and (source_type == 'id') and (target_type == 'id')):
            return [self.d_orth_human_to_mouse.get(g, 'na') for g in genes]
        
        elif ((source_organism == 'human') and (target_organism == 'mouse') 
           and (source_type == 'name') and (target_type == 'id')):
            genes = [self.d_human_name_to_id.get(g, 'na') for g in genes]
            return [self.d_orth_human_to_mouse.get(g, 'na') for g in genes]
        
        elif ((source_organism == 'human') and (target_organism == 'mouse') 
           and (source_type == 'id') and (target_type == 'name')):
            genes = [self.d_orth_human_to_mouse.get(g, 'na') for g in genes]
            return [self.d_mouse_id_to_name.get(g, 'na') for g in genes]
        
        elif ((source_organism == 'human') and (target_organism == 'mouse') 
           and (source_type == 'name') and (target_type == 'name')):
            genes = [self.d_human_name_to_id.get(g, 'na') for g in genes]
            genes = [self.d_orth_human_to_mouse.get(g, 'na') for g in genes]
            return [self.d_mouse_id_to_name.get(g, 'na') for g in genes]
        
def standardize_adata(adata, standard_genes=None):
    '''Standardize an AnnData object to a set of standard genes.
    Args:
        adata (AnnData): AnnData object to be standardized. The adata.var.index
                            should be Ensembl IDs.
        standard_genes (list): List of standard genes to be used for standardization.
    '''
    if None is standard_genes:
        standard_gene_csv = pkg_resources.resource_filename('scmg',
                            'data/standard_genes.csv')
        standard_genes = pd.read_csv(standard_gene_csv)['human_id'].values

    if scipy.sparse.issparse(adata.X):
        adata.X = adata.X.toarray()

    current_genes = list(adata.var.index)

    X = np.zeros((adata.shape[0], len(standard_genes)))
    current_indices = []
    standard_indices = []

    for i, g in enumerate(standard_genes):
        if g in current_genes:
            current_indices.append(current_genes.index(g))
            standard_indices.append(i)

    X[:, standard_indices] = adata.X[:, current_indices]

    adata_standard = anndata.AnnData(
        X=X,
        obs=adata.obs.copy(),
        var=pd.DataFrame(index=standard_genes),
    )
    return adata_standard


import numpy as np
import pandas as pd
from sklearn.metrics.pairwise import pairwise_distances
from sklearn.neighbors import NearestNeighbors


class CellTypeSearcher:
    def __init__(self, ref_cell_emb_adata, emb_key='X_scmg'):
        self.ref_cell_emb_adata = ref_cell_emb_adata
        if emb_key not in ref_cell_emb_adata.obsm.keys():
            self.ref_emb = ref_cell_emb_adata.X
        else:
            self.ref_emb = ref_cell_emb_adata.obsm[emb_key]

        # Get the cell types and their counts
        self.cell_types, self.ct_counts = np.unique(
            ref_cell_emb_adata.obs['cell_type'], return_counts=True)

        # Create a matrix that for weighted maps from 
        # reference cells to cell types
        self.ref_cell_to_ct_mtx = np.zeros((ref_cell_emb_adata.shape[0], 
                                       len(self.cell_types)), dtype=np.float32)
        for i, ct in enumerate(self.cell_types):
            self.ref_cell_to_ct_mtx[ref_cell_emb_adata.obs['cell_type'] == ct, 
                               i] = 1 / self.ct_counts[i]

    def search_ref_cell(self, query_emb, project_umap=True, n_jobs=32):
        '''Search for the reference cells that are most similar to
        each query cell.
        Args:
            query_emb (np.array): query embedding of shape (n_cells, n_features)
        '''
        neigh = NearestNeighbors(n_neighbors=1, n_jobs=n_jobs)
        neigh.fit(self.ref_emb)

        # Find the nearest reference cell for each query cell
        dist, ind = neigh.kneighbors(query_emb)
        ind = ind.flatten()
        dist = dist.flatten()

        neighbor_df = pd.DataFrame({
            'ref_cell': self.ref_cell_emb_adata.obs.index[ind],
            'distance': dist,
        })

        if project_umap:
            neighbor_df['umap_x'] = self.ref_cell_emb_adata.obsm[
                                    'X_umap'][ind, 0]
            neighbor_df['umap_y'] = self.ref_cell_emb_adata.obsm[
                                    'X_umap'][ind, 1]

        return neighbor_df

    def search_ref_cell_types(self, query_emb, radius=2, n_jobs=32):
        '''Search for reference cell types that are most similar to 
        the query cells
        Args:
            query_emb (np.array): query embedding of shape (n_cells, n_features)
            radius (float): soft search radius
        '''
        # Calculate the pairwise distances between the query and reference cells
        q_to_r_dist_mtx = pairwise_distances(query_emb, self.ref_emb,
                                    metric='euclidean', n_jobs=n_jobs)

        # Calculate the weights bewteen the query and reference cells
        assert(radius > 0)
        q_to_r_weights = np.exp(-(q_to_r_dist_mtx / radius) ** 2)

        # Calculate the weighted maps from reference cells to cell types
        q_to_ct_mtx = q_to_r_weights @ self.ref_cell_to_ct_mtx

        # Calculate the weights of cell types
        ct_weights = q_to_ct_mtx.mean(axis=0)

        return pd.DataFrame(ct_weights, index=self.cell_types, 
                columns=['weight']).sort_values('weight', ascending=False)
